#!/usr/bin/python
#################
#
# bacula-server.py - workload test for bacula server
#
#   test_0##_* - Initialization & verification
#   test_1##_* - Localized stress tests
#   test_2##_* - Client/Server stress tests (requires multiple systems)
#

import unittest, os, sys, tempfile
import subprocess, testlib 
import socket, time 
from string import Template

class InitBacula(testlib.TestlibCase):

    def setUp(self):
        '''Set up prior to each test_* function'''
        self.fs_dir = os.path.abspath('.')
        self.tmpldir = self.fs_dir + "/templates"
        self.hostname = socket.gethostname()
        # Prepare for per-test teardowns
        self.teardowns = []

    def tearDown(self):
        '''Clean up after each test_* function'''
        os.chdir(self.fs_dir)

        # Handle per-test teardowns
        for func in self.teardowns:
            func()

    def test_000_debconf(self):
        '''Set debconf settings prior to installing mysql '''
        tmplfile = open(self.tmpldir + "/debconf-set.tmpl")
        debconffile = tempfile.NamedTemporaryFile()
        debconffile.write(tmplfile.read().replace("HOSTNAME", self.hostname));
        debconffile.flush()
        self.assertShellExitEquals(0, ["debconf-set-selections", debconffile.name])

    def test_010_install(self):
        '''Install Bacula for testing '''
        # by installing mysql-server separately from bacula, we avoid bug LP:760443
        Package_list = ['mysql-server', 'bacula-director-mysql', 'bacula-sd-mysql', 'bacula-fd', 'bacula-console']
        for pkg in Package_list:
            self.assertShellExitEquals(0, ["apt-get","-y","install", pkg ])

    def test_020_build_conf_files(self):
        '''Setup Bacula configuration Files'''
        # Read in the default passwords
        passwd = {}
        for line in open('/etc/bacula/common_default_passwords'):
            if "=" not in line : continue
            var, val = line.split('=')
            passwd[var] = val.rstrip()
        passwd['HOSTNAME'] = self.hostname
        
        # Bacula SD
        sd_conf = open('/etc/bacula/bacula-sd.conf', 'w')
        for line in open(self.tmpldir + '/bacula-sd-conf.tmpl'):
            sd_conf.write(Template(line).substitute(passwd))
        sd_conf.close() 

        # Bacula FD
        fd_conf = open('/etc/bacula/bacula-fd.conf', 'w')
        for line in open(self.tmpldir + '/bacula-fd-conf.tmpl'):
            fd_conf.write(Template(line).substitute(passwd))
        fd_conf.close() 

        # Bacula Director
        dir_conf = open('/etc/bacula/bacula-dir.conf', 'w')
        for line in open(self.tmpldir + '/bacula-dir-conf.tmpl'):
            dir_conf.write(Template(line).substitute(passwd))
        dir_conf.close() 

        # Bacula Console
        bcon_conf = open('/etc/bacula/bconsole.conf', 'w')
        for line in open(self.tmpldir + '/bconsole-conf.tmpl'):
            bcon_conf.write(Template(line).substitute(passwd))
        bcon_conf.close() 

        # Make sure bacula has permissions to the backup area and a blank file
        self.assertShellExitEquals(0,["chown", "bacula:bacula", "/srv"])
        self.assertShellExitEquals(0,["sudo", "-u", "bacula", "touch", "/srv/backup"])
  

    def test_021_bacula_fd(self):
        '''Setup Bacula File Daemon'''
        self.assertShellExitEquals(0, ["bacula-fd","-t"])
        self.assertShellExitEquals(0, ["service", "bacula-fd","restart"])

    def test_022_bacula_sd(self):
        '''Setup Bacula SD'''
        self.assertShellExitEquals(0, ["bacula-sd","-t"])
        self.assertShellExitEquals(0, ["service", "bacula-sd","restart"])

    def test_022_bacula_dir(self):
        '''Setup Bacula Director'''
        self.assertShellExitEquals(0, ["bacula-dir","-t"])
        self.assertShellExitEquals(0, ["service", "bacula-director","restart"])

    def test_023_bacula_console(self):
        '''Setup Bacula Console'''
        self.assertShellExitEquals(0, ["bconsole","-t"])

    def test_030_basic_smoke(self):
        '''Smoketest'''
        bconcmd = 'echo "run job=Catalog yes" | bconsole'
        self.assertShellExitEquals(0,["sh", "-c", bconcmd])
        time.sleep(5)
        bconcmd = 'echo "message" |bconsole |tee bacula.log'
        (rc, report) = testlib.cmd(["sh", "-c", bconcmd])
        expected = 0
        result = 'Got exit code %d, expected %d\n' % (rc, expected)
        self.assertEquals(expected, rc, result + report)
        result = "Error in the setup\n"
        self.assertEquals(report.find("Backup Error"), -1, result + report)

if __name__ == '__main__':
    testlib.require_sudo()
    unittest.main(verbosity=2, failfast=True)



