#! /usr/bin/python -B
#-*- coding: utf-8 -*-

import logging
from helpers import run, transfer, install, remove, add_ppa

class Cassandra():
    def __init__(self, ppa=None, version=None, servers=None):
        if ppa:
            self.ppa = ppa
        else:
            self.ppa = "deb http://www.apache.org/dist/cassandra/debian %sx main" \
                    % version

        self.servers = servers

    def setup(self):
        from time import sleep

        logging.debug("Start Cassandra:setup")
        logging.info("Setting up Cassandra")
        add_ppa("cassandra", self.ppa, self.servers)
        install("cassandra", self.servers, "--allow-unauthenticated")
        logging.info("Waiting 15 seconds for cassandra to settle")
        sleep(15)
        logging.debug("End Cassandra:setup")

    def ycsb_setup(self):
        logging.debug("Start Cassandra:ycsb_setup")
        logging.info("Initializing Cassandra for YCSB")
        if self.servers:
            transfer("ycsb_init", self.servers)

        run("cassandra-cli -h localhost -f ycsb_init", self.servers)
        logging.debug("End Cassandra:ycsb_setup")

    def ycsb_cleanup(self):
        logging.debug("Start Cassandra:ycsb_cleanup")
        logging.info("Cleaning up Cassandra YCSB DB")
        if self.servers:
            transfer("ycsb_cleanup", self.servers)

        run("cassandra-cli -h localhost -f ycsb_cleanup", self.servers)
        logging.debug("End Cassandra:ycsb_cleanup")

    def cleanup(self):
        logging.debug("Start Cassandra:cleanup")
        remove("cassandra", self.servers)
        run("sudo apt-get --purge -y autoremove", self.servers)
        run("sudo rm -f /etc/apt/sources.list.d/cassandra.list*", self.servers)
        run("sudo apt-get update", self.servers)
        run("sudo rm -rf /var/lib/cassandra /var/log/cassandra", self.servers)
        run("rm -f ycsb_*", self.servers)
        logging.debug("End Cassandra:cleanup")

def main(argv=None):
    from optparse import OptionParser, make_option

    usage = "Usage: %prog [options] <server> <server...n>"
    epilog = "If no arguments are given, cassandra will be setup on localhost. " + \
            "The <server> arguments are machines where cassandra will be installed."
    option_list = [
        make_option("-l", "--log-level", default="info", dest="log_level",
                    choices=["debug", "info", "warning", "error", "critical"],
                    help="[ debug | *info | warning | error | critical ]"),
        make_option("--cleanup", action="store_true", dest="cleanup",
                    help="reverts changes made by the test"),
        make_option("--remove", action="store_true", dest="remove",
                    help="reverts changes made by the test, but does not run the test"),
        make_option("--setup", action="store_true", dest="setup",
                    help="install and configure cassandra for ycsb testing"),
        make_option("--ycsb-clean", action="store_true",
                    dest="ycsb_clean",
                    help="cleans out the ycsb database data from cassandra"),
        make_option("-p", "--ppa", dest="ppa",
                    help="ppa:<path> or \"deb  <url>\""),
        make_option("-v", "--version", dest="version",
                    choices=["07", "08", "10", "11"],
                    default="10", help="uses the default apache.org repo: " \
                    "cassandra version [ 07 | 08 | 10 | 11 ]"),
    ]
    parser = OptionParser(usage=usage, epilog=epilog, option_list=option_list)
    (options, args) = parser.parse_args()

    logging.basicConfig(format='%(asctime)s %(levelname)s:%(message)s',
                        level=options.log_level.upper())

    if not options.remove and not options.setup and not options.ycsb_clean:
        parser.error("Requires an argument, run with -h to see " \
                     "available arguments")

    cassandra = Cassandra(options.ppa, options.version, args)
    if not options.remove and options.setup:
        cassandra.setup()
        cassandra.ycsb_setup()

    if options.ycsb_clean:
        cassandra.ycsb_clean()

    if options.remove or options.cleanup:
        cassandra.cleanup()

if __name__ == "__main__":
    import sys, atexit

    atexit.register(logging.shutdown)
    sys.exit(main(sys.argv[1:]))

