#! /usr/bin/python -B
#-*- coding: utf-8 -*-

import logging
from helpers import run, install, remove, add_ppa, del_ppa, get_hostname_ipaddr

def start_hadoop_services(servers=None):
    logging.debug('Begin start_hadoop_cluster')
    logging.info('Starting all hadoop services')
    run('sudo service hadoop-namenode start', servers)
    run('sudo service hadoop-datanode start', servers)
    run('sudo service hadoop-secondarynamenode start', servers)
    run('sudo service hadoop-jobtracker start', servers)
    run('sudo service hadoop-tasktracker start', servers)
    logging.debug('End start_hadoop_cluster')

def stop_hadoop_services(servers=None):
    logging.debug('Begin stop_hadoop_cluster')
    logging.info('Stopping all hadoop services')
    run('sudo service hadoop-jobtracker stop', servers, False)
    run('sudo service hadoop-tasktracker stop', servers, False)
    run('sudo service hadoop-namenode stop', servers, False)
    run('sudo service hadoop-datanode stop', servers, False)
    run('sudo service hadoop-secondarynamenode stop', servers, False)
    logging.debug('End stop_hadoop_cluster')

def configure_java(servers=None, with_oracle_jvm=False):
    if not with_oracle_jvm:
        install('openjdk-6-jre-zero', servers)
        run("sudo sed 's/HADOOP_OPTS=-server/ a\\" + \
            "export HADOOP_OPTS=-zero/' -i /etc/hadoop/conf/hadoop-env.sh",
            servers)

def format_hdfs(servers=None):
    run('sudo rm -Rf /var/lib/hadoop/cache/*', servers)
    run('sudo dpkg-reconfigure hadoop-conf-pseudo', servers)
    run("sudo su -c 'hadoop namenode -format' hdfs", servers)

def pseudo_cluster_setup(master=None, ppa=None, with_oracle_jvm=False):
    logging.debug('Begin pseudo_cluster_setup')
    if not ppa:
        ppa = 'ppa:hadoop-ubuntu/dev'
    add_ppa('hadoop', ppa, master)
    install('hadoop-conf-pseudo', master)
    stop_hadoop_services(master)
    configure_java(master, with_oracle_jvm)
    format_hdfs(master)
    start_hadoop_services(master)
    logging.debug('End pseudo_cluster_setup')

def multi_node_cluster_setup(master, slaves, ppa=None,
                             with_oracle_jvm=False):
    logging.debug('Begin multi_node_cluster_setup')
    logging.info('Configuring Hadoop master ' + master)
    if not ppa:
        ppa = 'ppa:hadoop-ubuntu/dev'
    add_ppa('hadoop', ppa, master)
    install('hadoop-conf-pseudo', master)

    stop_hadoop_services(master)
    configure_java(master, with_oracle_jvm)
    run('sudo echo %s > masters' % master, master)
    run('sudo mv -f masters /etc/hadoop/conf/masters', master)
    run('sudo echo %s > slaves' % master, master)
    for slave in slaves:
        run('sudo echo %s >> slaves' % slave, master)
    run('sudo mv -f slaves /etc/hadoop/conf/slaves', master)
    run("sudo sed -e 's|\(\s*<value>hdfs://\).*\(</value>\)|\\1" + \
        master + ":54310\\2|' " + \
        "-i /etc/hadoop/conf/core-site.xml", master)
    run("sudo sed -e 's|\(\s*<value>\).*\(</value>\)|\\1" + \
        master + ":54311\\2|' " + \
        "-i /etc/hadoop/conf/mapred-site.xml", master)
    run("sudo sed -e 's|\(\w*<value>\)1\(</value>\)|\\1" + \
        "%d\\2|' " % (len(slaves) + 1) + \
        "-i /etc/hadoop/conf/hdfs-site.xml", master)

    hostname, ipaddr = get_hostname_ipaddr(master)
    run('cp -f /etc/hosts hosts', master)
    run("sudo sed -e 's/127\.0\.1\.1.*" + hostname + "/" + \
        ipaddr + "\t" + hostname + "/' -i /etc/hosts", master)
    for slave in slaves:
        logging.debug('Adding %s to /etc/hosts on %s' % (slave, master))
        hostname, ipaddr = get_hostname_ipaddr(slave)
        run("sudo sed '/%s/ a\\%s\t%s' -i /etc/hosts" % \
            (master, ipaddr, hostname), master)

    logging.info('Configuring Hadoop slaves ' + ' '.join(slaves))
    add_ppa('hadoop', ppa, slaves)
    install('hadoop-conf-pseudo openjdk-6-jre-zero', slaves)
    stop_hadoop_services(slaves)
    configure_java(slaves, with_oracle_jvm)
    run("sudo sed -e 's|\(\s*<value>hdfs://\).*\(</value>\)|\\1" + \
        master + ":54310\\2|' " + \
        "-i /etc/hadoop/conf/core-site.xml", slaves)
    run("sudo sed -e 's|\(\s*<value>\).*\(</value>\)|\\1" + \
        master + ":54311\\2|' " + \
        "-i /etc/hadoop/conf/mapred-site.xml", slaves)
    run("sudo sed -e 's|\(\s*<value>\)\w*\(</value>\)|\\1" + \
        "%d\\2|' " % (len(slaves) + 1) + \
        "-i /etc/hadoop/conf/hdfs-site.xml", slaves)

    mhostname, mipaddr = get_hostname_ipaddr(master)
    for slave in slaves:
        logging.debug('Fixing /etc/hosts on ' + slave)
        run('cp -f /etc/hosts hosts', slave)
        hostname, ipaddr = get_hostname_ipaddr(slave)
        run("sudo sed -e " + "'s/127\.0\.1\.1.*" + hostname + "/" + \
            ipaddr + "\t" + hostname + "/' " + "-i /etc/hosts", slave)
        run("sudo sed '/%s/ a\\%s\t%s' -i /etc/hosts" % \
            (slave, mipaddr, mhostname), slave)

    format_hdfs(master)

    # start hdfs daemons
    run('sudo service hadoop-namenode start', master)
    run('sudo service hadoop-datanode start', slaves)
    run('sudo service hadoop-datanode start', master)
    run('sudo service hadoop-secondarynamenode start', master)

    # start mapreduce daemons
    run('sudo service hadoop-jobtracker start', master)
    run('sudo service hadoop-tasktracker start', slaves)
    run('sudo service hadoop-tasktracker start', master)
    logging.debug('End multi_node_cluster_setup')

def setup(args):
    from time import sleep

    logging.debug('Begin setup')
    if not args.slaves:
        logging.info('Setting up Hadoop (pseudo cluster config)')
        pseudo_cluster_setup(args.master, args.ppa,
                             args.with_oracle_jvm)
    else:
        logging.info('Setting up Hadoop (multi-node cluster config)')
        multi_node_cluster_setup(args.master, args.slaves, args.ppa,
                                 args.with_oracle_jvm)

    logging.info('Waiting 30 seconds for hadoop to settle')
    sleep(30)
    logging.debug('End setup')

def clean(args):
    logging.debug('Begin clean')
    remove('hadoop-conf-pseudo openjdk-6-jre-zero bigtop-utils',
           args.servers)
    run('sudo apt-get --purge -y autoremove && ' \
        'sudo rm -rf /var/lib/hadoop /var/log/hadoop ' \
        '/var/log/upstart/hadoop*', args.servers, False)
    run("sudo mv -f hosts /etc/hosts", args.servers, False)
    del_ppa('hadoop-ubuntu', args.servers)
    logging.debug('End clean')

def run_test(args):
    logging.info('Begin run_test ' + args.test + \
                 ' with test_args: ' + args.test_args)
    run('hadoop jar %s %s %s' % (args.jar, args.test, args.test_args),
         args.master, print_to_console=True)
    logging.info('End run_test ' + args.test + \
                 ' with test_args: ' + args.test_args)

def main(argv=None):
    import argparse

    description = 'Hadoop workload test script.'
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('--log-level', default='info',
                        choices=['debug','info','warning','error','critical'],
                        help='sets the logging level, default is info')
    subparsers = parser.add_subparsers()
    setup_parser = subparsers.add_parser('setup', help='install and configure hadoop')
    setup_parser.add_argument('--ppa', default='ppa:hadoop-ubuntu/dev',
                              metavar='ppa:<path> or "deb  <url>"',
                              help='ppa or deb url to install hadoop from')
    setup_parser.add_argument('--with-oracle-jvm', action='store_true',
                              help='use the oracle jvm')
    setup_parser.add_argument('--master', help='hadoop master')
    setup_parser.add_argument('--slaves', nargs='*', help='hadoop slaves')
    setup_parser.set_defaults(func=setup)
    clean_parser = subparsers.add_parser('clean',
                                         help='reverts changes made by setup')
    clean_parser.add_argument('servers', nargs='+',
                              help='list of servers to revert')
    clean_parser.set_defaults(func=clean)
    test_parser = subparsers.add_parser('test', help='runs a hadoop test')
    test_parser.add_argument('test', help='hadoop test name')
    test_parser.add_argument('--test-args', help='test arguments')
    test_parser.add_argument('--jar', help='path to test jar file')
    test_parser.add_argument('--master', help='hadoop master')
    test_parser.set_defaults(func=run_test)
    parsed_args = parser.parse_args()

    logging.basicConfig(format='%(asctime)s %(levelname)s:%(message)s',
                        level=parsed_args.log_level.upper())
    logging.debug(parsed_args)
    parsed_args.func(parsed_args)

if __name__ == '__main__':
    import sys, atexit

    atexit.register(logging.shutdown)
    sys.exit(main(sys.argv[1:]))

