import re
import sys
import logging
from subprocess import Popen, PIPE, STDOUT

def run(cmd, hosts=None, exit_on_error=True, print_to_console=False):
    log = "Running " + cmd
    if hosts:
        log = log + " on " + " ".join(__encap(hosts))
        cmd = "ssh -q %(host)s \"" + cmd + "\""

    logging.debug(log)
    return __run(cmd, hosts, exit_on_error, print_to_console)

def transfer(path, hosts, opts=None, exit_on_error=True,
             print_to_console=False):
    if not hosts:
        logging.critical("A host or list of hosts was not provided.")
        sys.exit(1)

    if opts:
        cmd = "scp -q %s %s" % (opts, path)
    else:
        cmd = "scp -q " + path

    cmd = cmd + " %(host)s:"
    log = "Transfering " + path + " to " + " ".join(__encap(hosts))
    logging.info(log)
    return __run(cmd, hosts, exit_on_error, print_to_console)

def install(packages, hosts=None, opts=None,
            exit_on_error=True):
    log = "Installing " + packages
    if hosts:
        log = log + " on " + " ".join(__encap(hosts))

    cmd = "sudo DEBIAN_FRONTEND=noninteractive apt-get install -y "
    if opts:
        cmd = cmd + opts + " "

    logging.info(log)
    return run(cmd + packages, hosts, exit_on_error)

def remove(packages, hosts=None, opts=None,
           exit_on_error=False):
    log = "Removing " + packages
    if hosts:
        log = log + " from " + " ".join(__encap(hosts))

    cmd = "sudo DEBIAN_FRONTEND=noninteractive apt-get purge --purge -y "
    if opts:
        cmd = cmd + opts + " "

    logging.info(log)
    return run(cmd + packages, hosts, exit_on_error)

def add_ppa(name, ppa, hosts=None, exit_on_error=True):
    log = "Adding ppa "
    if ppa.startswith("ppa:"): 
        log = log + ppa
        install("python-software-properties", hosts)
        cmd = "sudo apt-add-repository -y " + ppa
    else:
        log = log + name
        cmd = "cat > %s.list <<EOF" % name
        cmd = cmd + "\n" + ppa + "\nEOF"
        run(cmd, hosts, exit_on_error)
        cmd = "sudo mv %s.list /etc/apt/sources.list.d/" % name

    if hosts:
        log = log + " on " + " ".join(__encap(hosts))

    logging.info(log)
    return run(cmd + " && sudo apt-get update", hosts,
               exit_on_error)

def del_ppa(name, hosts=None, exit_on_error=False):
    log = "Deleting ppa " + name
    if hosts:
        log = log + " from " + " ".join(__encap(hosts))

    logging.info(log)
    return run("sudo rm -f /etc/apt/sources.list.d/" + name + "*.list*" + \
               " && sudo apt-get update", hosts, exit_on_error)

def get_hostname_ipaddr(addr):
    if re.search('[a-zA-Z]', addr):
        hostname = addr
        ipaddr = run("ip -f inet -o addr show eth0 | " + \
                     "sed -e 's|^.*inet \(.*\)/.*$|\\1|'",
                     addr)[addr].rstrip('\n')
    else:
        ipaddr = addr
        hostname = run('hostname', addr)[addr].rstrip('\n')
    return hostname, ipaddr

def __encap(val):
    if isinstance(val, str) or val == None:
        return [val]
    else:
        return val

def __run(cmd, hosts=None, exit_on_error=True, print_to_console=False):
    output = {}
    for host in __encap(hosts):
        if host:
            cur_cmd = cmd % {"host" : host}
        logging.debug("__run: " + cur_cmd)
        proc = Popen(cur_cmd, stdout=PIPE, stderr=STDOUT, shell=True)
        if print_to_console:
            for line in iter(proc.stdout.readline, ''):
                print line.rstrip('\n')
        output[host], stderr = proc.communicate()
        if exit_on_error and proc.returncode != 0:
            logging.error("Command failed: '" + cur_cmd + "'\n" + \
                          output[host])
            sys.exit(proc.returncode)
    return output

