#! /usr/bin/python -B
#-*- coding: utf-8 -*-

from os import path
import logging
from helpers import run, transfer, install, remove, add_ppa, del_ppa

class HBase():
    def __init__(self, master=None):
        self.master = master

    def setup(self, ppa=None):
        from time import sleep

        logging.debug("Begin HBase:setup")
        logging.info("Setting up HBase")
        if not ppa:
            ppa = "ppa:hadoop-ubuntu/dev"

        add_ppa("hadoop", ppa, self.master)
        install("hbase openjdk-6-jre-zero", self.master)

        logging.info("Configuring hbase")
        if self.master:
            transfer(path.dirname(path.realpath(__file__)) + \
                     "/hbase-pseudo-distributed-config", self.master)

        run("sh hbase-pseudo-distributed-config", self.master)
        logging.info("Waiting 30 seconds for hbase to settle")
        sleep(30)
        logging.debug("End HBase:setup")

    def clean(self):
        logging.debug("Begin HBase:clean")
        run("sudo rm -rf /var/lib/hbase /var/log/hbase", self.master)
        run("sudo rm -rf /etc/hadoop-zookeeper " + \
            "/etc/alternatives/hadoop-zookeeper-conf", self.master)
        remove("hbase openjdk-6-jre-zero", self.master)
        run("sudo apt-get --purge -y autoremove", self.master)
        del_ppa("hadoop-ubuntu", self.master)
        logging.debug("End HBase:clean")

    def ycsb_setup(self):
        logging.debug("Start HBase:ycsb_setup")
        logging.info("Initializing HBase for YCSB")
        if self.master:
            transfer(path.dirname(path.realpath(__file__)) + "/ycsb_init",
                     self.master)

        run("/usr/lib/hbase/bin/hbase shell ycsb_init", self.master)
        logging.debug("End HBase:ycsb_setup")

    def ycsb_clean(self):
        logging.debug("Start HBase:ycsb_clean")
        logging.info("Cleaning up HBase YCSB DB")
        if self.master:
            transfer(path.dirname(path.realpath(__file__)) + "/ycsb_clean",
                     self.master)

        run("/usr/lib/hbase/bin/hbase shell ycsb_clean", self.master)
        logging.debug("End HBase:ycsb_clean")

    def run_test(self, test, nclients, opts=None):
        logging.info("Begin HBase:run_test " + test)
        cmd = "hbase org.apache.hadoop.hbase.PerformanceEvaluation"
        if opts:
            cmd = cmd + " " + opts
        cmd = cmd + " " + test + " " + nclients
        run(cmd, self.master, print_to_console=True)
        logging.info("End HBase:run_test " + test)

def main(argv=None):
    from optparse import OptionParser, make_option

    usage = "Usage: %prog [options]"
    epilog = "If no arguments are given, hbase will be setup on localhost."
    option_list = [
        make_option("-l", "--log-level", default="info", dest="log_level",
                    choices=["debug", "info", "warning", "error", "critical"],
                    help="[ debug | *info | warning | error | critical ]"),
        make_option("--setup", action="store_true", dest="setup",
                    help="install and configure hbase"),
        make_option("--clean", action="store_true", dest="clean",
                    help="reverts changes made by the test"),
        make_option("--ycsb-setup", action="store_true", dest="ycsb_setup",
                    help="adds database entries for ycsb"),
        make_option("--ycsb-clean", action="store_true", dest="ycsb_clean",
                    help="removes database entries for ycsb"),
        make_option("-p", "--ppa", dest="ppa",
                    default="ppa:hadoop-ubuntu/dev",
                    help="ppa:<path> or \"deb  <url>\""),
        make_option("-m", "--master", dest="master",
                    help="hbase master"),
        make_option("-t", "--test", dest="test",
                    help="hbase performance evaluation tests"),
        make_option("-n", "--nclients", dest="nclients",
                   help="hbase performance evaluation number of clients"),
        make_option("-o", "--opts", dest="opts",
                   help="hbase performance evaluation options"),
    ]
    parser = OptionParser(usage=usage, epilog=epilog, option_list=option_list)
    (options, args) = parser.parse_args()

    logging.basicConfig(format='%(asctime)s %(levelname)s:%(message)s',
                        level=options.log_level.upper())

    if not options.clean and not options.setup and not \
       options.ycsb_setup and not options.ycsb_clean and not \
       options.test:
        parser.error("Requires an argument, run with -h to see " \
                     "available arguments")

    hbase = HBase(options.master)
    if options.setup:
        hbase.setup(options.ppa)

    if options.test and options.nclients:
        hbase.run_test(options.test, options.nclients, options.opts)

    if options.ycsb_setup:
        hbase.ycsb_setup()

    if options.ycsb_clean:
        hbase.ycsb_clean()

    if options.clean:
        hbase.clean()

if __name__ == "__main__":
    import sys, atexit

    atexit.register(logging.shutdown)
    sys.exit(main(sys.argv[1:]))

