#! /usr/bin/python -B
#-*- coding: utf-8 -*-

import logging
from helpers import run, install, remove

class YCSB:
    def __init__(self, package=None, clients=None):
        if not package:
            self.package = "ycsb-0.1.4-beta-0.tar.gz"
        else:
            self.package = package

        self.clients = clients

    def setup(self, url=None):
        from os.path import isfile

        logging.debug("Start YCSB:setup")
        logging.info("Setting up YCSB")
        if not url:
            self.url = "https://github.com/downloads/brianfrankcooper/YCSB/ycsb-0.1.4-beta-0.tar.gz"

        # Downloading the YCSB binaries
        if not isfile(self.package):
            logging.info("Downloading " + self.package)
            run("wget %s" % url, self.clients)

        # Install the YCSB binaries on the host machines
        logging.info("Extracing " + self.package)
        run("tar xf " + self.package, self.clients)

        # Installing YCSB dependencies (java)
        install("openjdk-6-jre-headless", self.clients)
        logging.debug("End YCSB:setup")

    def clean(self):
        logging.debug("Begin YCSB:clean")
        remove("openjdk-6-jre-headless", self.clients)
        run("sudo apt-get --purge -y autoremove", self.clients)
        run("rm -Rf ycsb*/ " + self.package, self.clients)
        logging.debug("End YCSB:clean")

    def __cassandra(self, workload, test_args, version=None):
        if not version:
            version = "10"
        cmd = "python ycsb-*/bin/ycsb load cassandra-%s ycsb-*/workloads/%s %s" % \
                (version, workload, test_args)
        logging.info("Running cassandra version %s, workload %s, with args: %s" \
                     % (version, workload, test_args))
        run(cmd, self.clients, print_to_console=True)

    def __hbase(self, workload, test_args):
        cmd = "python ycsb-*/bin/ycsb load hbase ycsb-*/workloads/%s %s" % \
                (workload, test_args)
        logging.info("Running hbase with workload %s and with test " \
                     "args: %s" % (workload, test_args))
        run(cmd, self.clients, print_to_console=True)

    def connector(self, database, workload, test_args, version=None):
        logging.debug("Start YCSB:connector")
        logging.info("Setting up YCSB connector %s" % database)
        if database == "cassandra":
            self.__cassandra(workload, test_args, version)
        elif database == "hbase":
            self.__hbase(workload, test_args)

        logging.debug("End YCSB:connector")

def main(argv=None):
    from optparse import OptionParser, make_option

    usage = "Usage: %prog [options] <client> <client...n>"
    epilog = "The <client> arguments specify where to install and " + \
            "run ycsb from."
    option_list = [
        make_option("-l", "--log-level", default="info", dest="log_level",
                    choices=["debug", "info", "warning", "error", "critical"],
                    help="[ debug | *info | warning | error | critical ]"),
        make_option("--setup", action="store_true", dest="setup",
                    help="install and configure ycsb"),
        make_option("--clean", action="store_true", dest="clean",
                    help="reverts changes made by the test"),
        make_option("-t", "--test", dest="test",
                    metavar="\"<args>\"",
                    help="ycsb connector arguments"),
        make_option("-p", "--package", dest="package", default="ycsb-0.1.4-beta-0.tar.gz",
                    help="tarball containing YCSB binaries (%default)"),
        make_option("-u", "--url", dest="url",
                    default="https://github.com/downloads/brianfrankcooper/YCSB/ycsb-0.1.4-beta-0.tar.gz",
                    help="URL to download the YCSB package from (%default)"),
        make_option("-d", "--database", dest="database", default="hbase",
                    choices=["cassandra","hbase"],
                    help="distributed database you want to test, i.e. %default"),
        make_option("-v", "--version", dest="version",
                    help="which version of the distributed data store to test (needed for cassandra)"),
        make_option("-w", "--workload", dest="workload", default="workloada",
                    choices=["workloada", "workloadb", "workloadc", "workloadd", "workloade",
                             "workloadf"],
                    help="which YCSB supplied workload to use: [workload<a, b, c, d, e, f>]"),
    ]
    parser = OptionParser(usage=usage, epilog=epilog, option_list=option_list)
    (options, args) = parser.parse_args()

    logging.basicConfig(format='%(asctime)s %(levelname)s:%(message)s',
                        level=options.log_level.upper())

    if not options.clean and not options.setup and not options.test:
        parser.error("Requires an argument, run with -h to see " \
                     "available arguments")

    ycsb = YCSB(options.package, args)
    if options.setup:
        ycsb.setup(options.url)

    if options.test:
        ycsb.connector(options.database, options.workload,
                       options.test, options.version)

    if options.clean:
        ycsb.clean()

if __name__ == "__main__":
    import sys, atexit

    atexit.register(logging.shutdown)
    sys.exit(main(sys.argv[1:]))

