#!/usr/bin/python
#################
#
#   Lamp stack workload testsuite
#
#   test_0##_* - Initialization & verification
#   test_1##_* - Localized stress tests
#   test_2##_* - Client/Server stress tests (requires multiple systems)
#

import unittest, os, sys
import subprocess, testlib 
import urllib2

class InitLampStack(testlib.TestlibCase):

    def setUp(self):
        '''Set up prior to each test_* function'''
        self.fs_dir = os.path.abspath('.')
        self.NumProcs = os.sysconf("SC_NPROCESSORS_ONLN")

        # Prepare for per-test teardowns
        self.teardowns = []

    def tearDown(self):
        '''Clean up after each test_* function'''
        os.chdir(self.fs_dir)

        # Handle per-test teardowns
        for func in self.teardowns:
            func()

    def test_000_debconf(self):
        '''Set debconf settings prior to installing mysql '''
        self.assertShellExitEquals(0, ["debconf-set-selections","debconf.set"])

    def test_010_install(self):
        '''Install lamp-server packages & siege for testing '''
        Package_list = ['lamp-server^', 'siege', 'python-mysqldb']
        for pkg in Package_list:
            self.assertShellExitEquals(0, ["apt-get","-y","install", pkg ])

    def test_020_php(self):
        '''Test the Apache2/PHP interface and install php script for data mining '''
        php_fh = open("/var/www/test.php", "w")
        php_fh.write("<? $world=\"Hello World!\"; echo $world; ?>")
        php_fh.close()
        fh = urllib2.urlopen('http://localhost/test.php')
        self.assertNotEquals(fh, None)
        output = fh.read()
        tFail = False
        try:
            self.assertEquals('Hello World!', output)
        except: # LP:953081 workaround
            tFail = True
            self.assertShellExitEquals(0,["sudo", "service", "apache2", "restart"])
            fh = urllib2.urlopen('http://localhost/test.php')
            self.assertNotEquals(fh, None)
            output = fh.read()
            self.assertEquals('Hello World!', output)
        self.assertShellExitEquals(0,["sudo", "cp", "psql.php", "/var/www/"])
        if tFail:
            self.fail(msg='LP:953081 detected')

    def test_030_initDB(self):
        '''Load the test data into MySQL'''
        os.chdir("test-db/employees_db/")
        self.assertShellExitEquals(0, ["mysql", "--user=root", "--password=ubuntu", "-t"], stdin=file("employees.sql"))

    def test_040_verifyDB(self):
        '''Verify database setup with md5sums'''
        os.chdir("test-db/employees_db/")
        (rc, report) = testlib.cmd(["mysql", "--user=root", "--password=ubuntu", "-t"], stdin=file('test_employees_md5.sql'))
        expected = 0
        result = 'Got exit code %d, expected %d\n' % (rc, expected)
        self.assertEquals(expected, rc, result + report)
        result = "Checksums didn't match the data in the InnoDB database\n"
        self.assertEquals(report.find("not ok"), -1, result + report)

    def test_100_apachebench(self):
        '''Apache Benchmark'''
        self.assertShellExitEquals(0, ["ab", "-n 100", "-c", str(self.NumProcs +4), 
             "http://localhost/psql.php"], stdout=file('ab.out', 'w'))
   
    def test_110_siege(self):
        '''Siege testing'''
        self.assertShellExitEquals(0, ["siege", "-b", "--log=siege.log", "-r 100 ", "-c",
             str(self.NumProcs +4), "http://localhost/psql.php"], stdout=file('siege.out', 'w'))
       
if __name__ == '__main__':
    testlib.require_sudo()
    unittest.main(verbosity=2)

