#!/bin/bash
#    Copyright (C) 2012 Canonical Ltd.
#    Author: Serge Hallyn <serge.hallyn@canonical.com>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License version 3,
#    as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Notes:
# 1. This should be rewritten as a part of lp:qa-regression-testing
# 2. run this as root

set -e

# determine architecture
# Code taken from debootstrap
if [ -x /usr/bin/dpkg ] && /usr/bin/dpkg --print-architecture >/dev/null 2>&1; then
    arch=`/usr/bin/dpkg --print-architecture`
elif type udpkg >/dev/null 2>&1 && udpkg --print-architecture >/dev/null 2>&1; then
    arch=`/usr/bin/udpkg --print-architecture`
else
    arch=$(arch)
    if [ "$arch" = "i686" ]; then
        arch="i386"
    elif [ "$arch" = "x86_64" ]; then
        arch="amd64"
    elif [ "$arch" = "armv7l" ]; then
        arch="armel"
    fi
fi

case "$arch" in
  armel) releases="oneiric precise" ;;
  armhf) releases="precise" ;;
  *)     releases="lucid oneiric precise" ;;
esac

apt-get install -y cgroup-lite libvirt-bin lxc lsb-release make gcc
hostrelease=`lsb_release -c | awk -F: '{ print $2 }'`
if [ $hostrelease != "precise" -a $hostrelease != "oneiric" ]; then
	echo "testsuite only supported on oneiric and precise"
fi

if [ $hostrelease = "precise" ] ; then
	echo "Running reboot test"
	make reboot-test || echo "FAILED to compile reboot test"
	./reboot-test || echo "reboot test FAILED"
fi

cat > /etc/lxc-test.conf << EOF
lxc.network.type=veth
lxc.network.link=virbr0
lxc.network.flags=up
EOF

if [ $hostrelease != "precise" ]; then
	for r in $releases; do
		rm -rf /var/cache/lxc/$r
	done
	cleanflag=""
else
	cleanflag="-F"
fi

for r in $releases; do
	c=${r}-test
	echo "Destroying any pre-existing container $c"
	lxc-destroy -n $c || true
	# create a container, install lxc (needed for lxc-execute to run),
	# and try lxc-execute
	echo "Creating container $c"
	lxc-create -f /etc/lxc-test.conf -t ubuntu -n $c -- -r $r ${cleanflag}
	if [ $? -ne 0 ]; then
		echo "FAIL creating container $c"
	fi
	echo "Chrooting in to install libcap2"
	chroot /var/lib/lxc/$c/rootfs apt-get -y install libcap2
	if [ $? -ne 0 ]; then
		echo "FAIL installing libcap2 in container $c"
	fi
	echo "Chrooting in to download lxc"
	chroot /var/lib/lxc/$c/rootfs apt-get -y install --download-only lxc
	if [ $? -ne 0 ]; then
		echo "FAIL downloading lxc in container $c"
	fi

	d=`mktemp -d`
	echo "extracting lxc for container $c into $d"
	dpkg -x /var/lib/lxc/$c/rootfs/var/cache/apt/archives/lxc_* $d
	if [ $? -ne 0 ]; then
		echo "FAIL extracting lxc for container $c"
	fi

	echo "copying /usr/lib/lxc into container"
	cp -a $d/usr/lib/lxc /var/lib/lxc/$c/rootfs/usr/lib/
	if [ $? -ne 0 ]; then
		echo "FAIL extracting lxc for container $c"
	fi

	rm -rf $d
	if [ $? -ne 0 ]; then
		echo "FAIL removing tmpdir $d"
	fi

	echo "Executing /bin/true on container $c"
	lxc-execute -n $c -- /bin/true
	if [ $? -ne 0 ]; then
		echo "FAIL running /bin/true in container $c"
	fi

	echo "Destroying container $c"
	lxc-destroy -n $c
	if [ $? -ne 0 ]; then
		echo "FAIL destroying $c"
	fi

	# re-create the container, testing the /var/cache/lxc/ this time
	lxc-create -f /etc/lxc-test.conf -t ubuntu -n $c -- -r $r
	lxc-start -n $c -d
	sleep 10s
	echo "Using lxc-ls to check for the running container $c"
	lxc-ls | tail -n +2 | grep -q $c   # if not found, error out
	if [ $? -ne 0 ]; then
		echo "FAILED lxc-ls"
	fi
	echo "Using lxc-ps to check for /sbin/init in container $c"
	lxc-ps -n $c -- -ef | grep -q '/sbin/init' # no init, error out
	if [ $? -ne 0 ]; then
		echo "FAILED lxc-ls"
	fi
	# should we try to ssh in?
	echo "stopping container $c"
	lxc-stop -n $c
	if [ $? -ne 0 ]; then
		echo "FAIL stopping container $c"
	fi
	lxc-destroy -n $c
	echo "destroying container $c"
	if [ $? -ne 0 ]; then
		echo "FAIL destroying container $c"
	fi
done

# other things to test:
# lxc-freeze; lxc-unfreeze; lxc-ps; lxc-console; lxc-clone
# lxc-cgroup; lxc-start-ephemeral
# lxc-start with console  (use expect?  or wait for a port to
# qa-regression-testing which has good support for that)

# if on precise, then also create a container with lxcbr0
#if [ `lsb_release -c | awk -F: '{ print $2 }'` = "precise" ]; then
#	for r in $releases; do
#		lxc-create -t ubuntu -n ${r}-1 -- -r $r
#	done
#fi

rm -f /etc/lxc-test.conf

exit 0
