#!/usr/bin/python
"""
osu_tests.py

Script to download and run OSU benchmark tests.
"""

import argparse
import os
import subprocess
import sys

# Default source and max processes
SOURCE = "http://mvapich.cse.ohio-state.edu/benchmarks/osu-micro-benchmarks-3.6.tar.gz"
DIR = "osu-micro-benchmarks-3.6"
MAX_PROCESSES = 2

# Packages to install from APT
PACKAGES = ["openmpi-bin", "openmpi-dev"]

# Tests that require exactly 2 processes
TWO_TESTS = ["osu_acc_latency",
        "osu_bibw",
        "osu_bw",
        "osu_get_bw",
        "osu_get_latency",
        "osu_latency",
        "osu_latency_mt",      # This test fails without multithread support
        "osu_passive_acc_latency",
        "osu_passive_get_bw",
        "osu_passive_get_latency",
        "osu_passive_put_bw",
        "osu_passive_put_latency",
        "osu_put_bibw",
        "osu_put_bw",
        "osu_put_latency"]

# Tests that can use 2 or more processes
ANY_TESTS = ["osu_allgather",
        "osu_allgatherv",
        "osu_allreduce",
        "osu_alltoall",
        "osu_barrier",
        "osu_bcast",
        "osu_gather",
        "osu_gatherv",
        "osu_mbw_mr",
        "osu_multi_lat",
        "osu_reduce",
        "osu_reduce_scatter",
        "osu_scatter",
        "osu_scatterv"]

def main():
    # Parse arguments
    parser = argparse.ArgumentParser(
            description="Run the OSU benchmark suite to test MPI")
    parser.add_argument("--hostfile", default=None,
            help="Specify a hostfile to use")
    parser.add_argument("--install", action="store_true",
            help="Install the necessary MPI packages on all hosts")
    parser.add_argument("-n", type=int, default=MAX_PROCESSES,
            help="Max number of processes to spawn")
    args = parser.parse_args()
    
    # Get a list of hosts
    hosts = get_hosts(args.hostfile)
    
    # Install the necessary MPI packages
    if args.install:
        for host in hosts:
            install_packages(host)
    
    # Install the OSU benchmark suite
    install_source()
    
    # Run two-process tests
    failed_tests = []
    for test in TWO_TESTS:
        print "Running test: " + test
        if run_test(DIR + "/" + test, 2, args.hostfile) == 0:
            print "Test passed\n"
        else:
            print "ERROR: Test failed\n"
            failed_tests.append(test)

    # Run any-process tests
    for test in ANY_TESTS:
        print "Running test: " + test
        if run_test(DIR + "/" + test, args.n, args.hostfile) == 0:
            print "Test passed\n"
        else:
            print "ERROR: Test failed\n"
            failed_tests.append(test)

    # Print final status
    if len(failed_tests) == 0:
        print "All tests passed.\n"
        return 0
    else:
        print "ERROR: The following tests failed"
        for test in failed_tests:
            print test
        print
        return 1

def get_hosts(hostfile):
    """
    Get a list of hosts from the hostfile
    """
    hosts = []
    
    if hostfile != None:
        fin = open(hostfile)
        for line in fin:
            hosts.append(line.split()[0])
        fin.close()
    else:
        hosts.append("localhost")
    
    return hosts

def install_packages(host):
    """
    Install PACKAGES on the specified host
    """
    command = ""
    if host != "localhost":
        command = "ssh -t " + host + " "
    command += "sudo apt-get install -y"
    for package in PACKAGES:
        command += " " + package
    
    return subprocess.call(command.split())

def install_source():
    """
    Download and build the source
    """
    tarball = SOURCE.split("/")[-1]
    
    # Download source
    command = "wget " + SOURCE
    command += " -O " + tarball
    if subprocess.call(command.split()):
        return False
    
    # Extract
    command = "tar -xf " + SOURCE.split("/")[-1]
    if subprocess.call(command.split()):
        return False
    
    # Build
    origin_dir = os.getcwd()
    os.chdir(origin_dir + "/" + DIR)
    if subprocess.call(["./configure"]):
        return False
    if subprocess.call(["make"]):
        return False
    os.chdir(origin_dir)
    
    return True

def run_test(test, n, hostfile):
    """
    Spawn an instance of MPI to run the specified test
    """
    command = "mpirun"
    command += " -n " + str(n)
    if hostfile != None:
        command += " --hostfile " + hostfile
    command += " " + test
    print command
    return subprocess.call(command.split())

if __name__ == "__main__":
    sys.exit(main())
