#!/usr/bin/python

import argparse
import bmark
import bmark.BmarkDirectory
import bmark.parser
import HTMLParser
import os
import os.path
import stat
import string
import subprocess
import sys
import time

# NB: this is the jenkinsapi module, not the module from
# the python-jenkins package which is called 'jenkins'.  It
# does not appear to be packaged yet but can be installed
# with:
#
#   # easy_install jenkinsapi
#
from jenkinsapi import jenkins

class MLStripper(HTMLParser.HTMLParser):
    def __init__(self):
        self.reset()
        self.fed = []

    def handle_data(self, d):
        self.fed.append(d)

    def get_data(self):
        d = ''.join(self.fed)
        n = string.find(d, 'Building on ')
        if n < 0:
            n = string.find(d, 'Building remotely on ')
        if n < 0:
            n = 0
        return d[n:]


def strip_tags(html):
    s = MLStripper()
    s.feed(html)
    return s.get_data()


class WantedJobs:

    def __init__(self, joblist):
        self.joblist = joblist
        self.wanted = {}
        self.get_wanted()
        return

    def get_wanted(self):
        if os.path.exists(self.joblist):
            fd = open(self.joblist, 'r+')
            line = fd.readline()
            data = line.strip(' \t\n').split()
            while len(data):
	        if data[0][0] == '#':
                   line = fd.readline()
                   data = line.strip(' \t\n').split()
                   continue
	        if data[0] not in self.wanted:
	            self.wanted[data[0]] = data[1:]
                line = fd.readline()
                data = line.strip(' \t\n').split()
            fd.close()

    def is_wanted(self, job_name):
	return job_name in self.wanted

    def get_bmark_name(self, job_name):
	return self.wanted[job_name][0]		# always the first element

    def get_parser_name(self, job_name):
	return self.wanted[job_name][1]		# always the second element


def main():
    #-- figure out the command line parameters
    home = os.getenv('HOME') or '/var/lib/jenkins'
    parser = argparse.ArgumentParser()
    parser.add_argument('--jobs', default='./jobs',
                        help='list of benchmark job names')
    parser.add_argument('--results', default='./results',
                        help='directory for storing benchmark results')
    parser.add_argument('--jenkins-username', default='jenkins',
                        help='login name for jenkins instance')
    parser.add_argument('--jenkins-password', default='jenkins',
                        help='password for the jenkins users')
    parser.add_argument('--url', default='http://tito/jenkins/',
                        help='location of jenkins instance')
    args = parser.parse_args()

    #-- figure out which jobs we actually want data for
    wanted_jobs = WantedJobs(args.jobs)

    #-- connect to the jenkins instance
    jenkins_connection = jenkins.Jenkins(args.url,
                                         args.jenkins_username,
                                         args.jenkins_password)

    #-- connect to the benchmark data collection
    bmd = bmark.BmarkDirectory.BmarkDirectory(args.results)

    #-- sort through all of the available job data
    jobs_list = jenkins_connection.get_jobs_list()
    sys.stderr.write('=== jobs list===\n')
    sys.stderr.flush()
    jobs_list.sort()
    for ii in jobs_list:
	if not wanted_jobs.is_wanted(ii):
	    continue
        sys.stderr.write('jobname: %s\n' % ii)
        sys.stderr.flush()
	job = jenkins_connection.get_job(ii)
	builds = job.get_build_dict()
	blist = builds.keys()
	blist.sort()
	for jj in blist:
            bld = job.get_build(jj)
	    if bld.is_good():
	        sys.stderr.write('    build %d...' % bld.id())
                sys.stderr.flush()

		#-- add the bmark data if we don't already have it
	        sys.stderr.write('get log...')
                sys.stderr.flush()
                cmd = ['wget', '-q', '-O', '-', '%sconsole' % builds[jj]]
	        p = subprocess.Popen(cmd, stdout=subprocess.PIPE)
                log, errs = p.communicate()
                timestamp = bld.get_timestamp()
	        bmark_name = wanted_jobs.get_bmark_name(ii)
		bmark_entry = str(timestamp)
                if not bmd.entry_exists(bmark_name, bmark_entry):
                    bmd.add_entry(bmark_name, bmark_entry)
		if bmd.entry_blob_exists(bmark_name, bmark_entry, 'raw_log'):
	            sys.stderr.write('ignoring duplicate raw_log, done.\n')
                    sys.stderr.flush()
		    continue
		slog = strip_tags(log)
		bmd.add_entry_blob(bmark_name, bmark_entry, 'raw_log', slog)

		#-- fill in the details of this data set
	        sys.stderr.write('parse...')
                sys.stderr.flush()

		dataset = bmark.parser.get_dataset_class(
					  	wanted_jobs.get_parser_name(ii))
		dataset.set_bmark_name(bmark_name)
		dataset.set_bmc(bmd)
		dataset.set_duration(bld.get_duration())
		dataset.set_timestamp(bld.get_timestamp())
		dataset.set_report_url(bld.get_result_url())
		dataset.set_common_values(slog)
		dataset.parse_log(slog)
		if bmd.entry_blob_exists(bmark_name, bmark_entry, 'dataset'):
	            sys.stderr.write('ignoring duplicate dataset, done.\n')
                    sys.stderr.flush()
		    continue
		bmd.add_entry_blob(bmark_name, bmark_entry, 'dataset',
			           str(dataset))
		
	        sys.stderr.write('done.\n')
                sys.stderr.flush()

if __name__ == '__main__':
    main()

