#!/usr/bin/python

import argparse
import hwpool
import importlib
import os
import os.path
import socket
import subprocess
import sys
import time
import tempfile
import yaml


class HardwarePoolNotFoundError(RuntimeError):
    pass


def get_machines_to_instantiate(pool, test_definition, tags):
    available_machines = pool.get_available_machines(tags)
    machines_to_instantiate = []
    machine_classes = test_definition['machine_classes']
    for machine_class_name, machine_class in machine_classes.iteritems():
        machines_needed = machine_class['machines']
        if machines_needed > len(available_machines):
            raise hwpool.InstantiationError('Not enough machines available')

        machines_to_use = available_machines[:machines_needed]
        available_machines = available_machines[machines_needed:]

        os.environ[machine_class_name.upper()] = ' '.join(machines_to_use)
        machines_to_instantiate.extend(machines_to_use)

    return machines_to_instantiate


def prepare_test(pool, test_definition, tags, keywords):
    machines_to_instantiate = get_machines_to_instantiate(pool,
                                                          test_definition,
                                                          tags)
    pool.instantiate_machines(machines_to_instantiate, keywords)


def get_cpu():
    result = 'unknown'
    count = 0
    fd = open('/proc/cpuinfo', 'r')
    line = fd.readline()
    while line:
        data = line.split(':')
        if len(data) >= 2:
            if data[0].strip() == 'model name':
                if result == 'unknown':
                    result = ' '.join(data[1:])
                    count = 1
                else:
                    count += 1
        line = fd.readline()
    fd.close()
    return (result.strip(), count)


def get_ram():
    result = 'unknown'
    fd = open('/proc/meminfo', 'r')
    line = fd.readline()
    while line:
        data = line.split(':')
        if len(data) >= 2:
            if data[0] == 'MemTotal':
                result = ' '.join(data[1:])
                break
        line = fd.readline()
    fd.close()
    return result.strip()


def print_test_machine_info(tags, keywords):
    msg = ''
    host = socket.gethostname()
    msg += 'RUNNER_HOST_NAME: ' + host + '\n'
    msg += 'RUNNER_HOST_IP: ' + socket.gethostbyname(host) + '\n'

    (sysname, nodename, release, version, machine) = os.uname()
    msg += 'RUNNER_HOST_UNAME: ' + ' '.join([sysname, nodename, release,
                                             version, machine,
                                            ]) + '\n'
    msg += 'RUNNER_HOST_ARCH: ' + machine + '\n'

    (cpu, cpu_count) = get_cpu()
    msg += 'RUNNER_HOST_CPU: ' + cpu + '\n'
    msg += 'RUNNER_HOST_CPU_COUNT: ' + str(cpu_count) + '\n'

    msg += 'RUNNER_HOST_RAM: ' + get_ram() + '\n'

    if tags == None:
       tags = ['None']
    msg += 'RUNNER_TAGS: ' + ' '.join(tags) + '\n'

    if keywords == None:
       keywords = ['None']
    msg += 'RUNNER_KEYWORDS: ' + ' '.join(keywords) + '\n'

    sys.stdout.write(msg)
    sys.stdout.flush()


def mark_run_test_start():
    msg = 'RUNNER_BMARK_START: '
    msg += '%s' % str(int(time.time()))
    msg += '\n'
    sys.stdout.write(msg)
    sys.stdout.flush()


def mark_run_test_stop():
    msg = 'RUNNER_BMARK_STOP: '
    msg += '%s' % str(int(time.time()))
    msg += '\n'
    sys.stdout.write(msg)
    sys.stdout.flush()


def run_test(test_definition):
    mark_run_test_start()
    executable = os.path.join(os.getcwd(), test_definition['executable'])
    subprocess.call([executable])
    mark_run_test_stop()


def main():
    # create the argument parser
    parser = argparse.ArgumentParser()

    # arguments used by everyone
    parser.add_argument('--check-if-possible', action='store_true')
    parser.add_argument('--tags', nargs='+', default='auto')
    parser.add_argument('--keywords', nargs='+')
    parser.add_argument('--hwpool', default='PandaKit')

    # arguments needed only by the JenkinsAPI allocator
    parser.add_argument('--jenkins-username', default='jenkins')
    parser.add_argument('--jenkins-password', default='jenkins')
    parser.add_argument('--jenkins-url', default='http://tito/jenkins/')

    # the final unnamed argument on the line
    parser.add_argument('test_definition_file')

    # now parse the arguments
    args = parser.parse_args()

    # Load the test definition file
    with open(args.test_definition_file, 'r') as f:
        test_definition = yaml.load(f.read())

    # Load the module that provides the hardware pool
    try:
        hwpool_module = importlib.import_module('%s.%s' %
                                                ('hwpool', args.hwpool))
    except ImportError:
        raise HardwarePoolNotFoundError(args.hwpool)
    try:
        hwpool_class = hwpool.hwpool
    except NameError:
        raise HardwarePoolNotFoundError(args.hwpool)
    pool = hwpool_class(args)

    # Change working directory to the directory the test definition file is in
    test_dir = os.path.dirname(args.test_definition_file)
    if test_dir:
        os.chdir(test_dir)

    if args.check_if_possible:
        try:
            get_machines_to_instantiate(pool, test_definition, args.tags)
        except hwpool.InstantiationError:
            return 1
        else:
            return 0
    else:
        # Run the test
        prepare_test(pool, test_definition, args.tags, args.keywords)
        print_test_machine_info(args.tags, args.keywords)
        run_test(test_definition)


if __name__ == '__main__':
    sys.exit(main())
